"""Models for root APIs."""

from dataclasses import dataclass
from enum import StrEnum

from .base import ResponseData

# --- ENUMS ----


class HostFeature(StrEnum):
    """HostFeature type.

    This is an incomplete list. Supervisor regularly adds support for new host
    features as users request them. Therefore when returning a list of host
    features, there may be some which are not in this list parsed as strings on
    older versions of the client.
    """

    DISK = "disk"
    HAOS = "haos"
    HOSTNAME = "hostname"
    JOURNAL = "journal"
    MOUNT = "mount"
    NETWORK = "network"
    OS_AGENT = "os_agent"
    REBOOT = "reboot"
    RESOLVED = "resolved"
    SERVICES = "services"
    SHUTDOWN = "shutdown"
    TIMEDATE = "timedate"


class SupervisorState(StrEnum):
    """SupervisorState type."""

    INITIALIZE = "initialize"
    SETUP = "setup"
    STARTUP = "startup"
    RUNNING = "running"
    FREEZE = "freeze"
    SHUTDOWN = "shutdown"
    STOPPING = "stopping"
    CLOSE = "close"


class UpdateChannel(StrEnum):
    """UpdateChannel type."""

    STABLE = "stable"
    BETA = "beta"
    DEV = "dev"


class LogLevel(StrEnum):
    """LogLevel type."""

    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


class UpdateType(StrEnum):
    """UpdateType type."""

    ADDON = "addon"
    CORE = "core"
    OS = "os"
    SUPERVISOR = "supervisor"


# --- OBJECTS ----


@dataclass(frozen=True, slots=True)
class RootInfo(ResponseData):
    """Root info object."""

    supervisor: str
    homeassistant: str | None
    hassos: str | None
    docker: str
    hostname: str | None
    operating_system: str | None
    features: list[HostFeature | str]
    machine: str | None
    machine_id: str | None
    arch: str
    state: SupervisorState
    supported_arch: list[str]
    supported: bool
    channel: UpdateChannel
    logging: LogLevel
    timezone: str


@dataclass(frozen=True, slots=True)
class AvailableUpdate(ResponseData):
    """AvailableUpdate type."""

    update_type: UpdateType
    panel_path: str
    version_latest: str
    name: str | None = None
    icon: str | None = None


@dataclass(frozen=True, slots=True)
class AvailableUpdates(ResponseData):
    """AvailableUpdates type."""

    available_updates: list[AvailableUpdate]



"""Models for resolution center APIs."""

from dataclasses import dataclass
from enum import StrEnum
from uuid import UUID

from .base import Options, ResponseData

# --- ENUMS ----


class SuggestionType(StrEnum):
    """SuggestionType type.

    This is an incomplete list. Supervisor regularly adds new types of suggestions as
    they are discovered. Therefore when returning a suggestion, it may have a type that
    is not in this list parsed as strings on older versions of the client.
    """

    ADOPT_DATA_DISK = "adopt_data_disk"
    CLEAR_FULL_BACKUP = "clear_full_backup"
    CREATE_FULL_BACKUP = "create_full_backup"
    DISABLE_BOOT = "disable_boot"
    ENABLE_NTP = "enable_ntp"
    EXECUTE_REBOOT = "execute_reboot"
    EXECUTE_REBUILD = "execute_rebuild"
    EXECUTE_RELOAD = "execute_reload"
    EXECUTE_REMOVE = "execute_remove"
    EXECUTE_REPAIR = "execute_repair"
    EXECUTE_RESET = "execute_reset"
    EXECUTE_RESTART = "execute_restart"
    EXECUTE_START = "execute_start"
    EXECUTE_STOP = "execute_stop"
    EXECUTE_UPDATE = "execute_update"
    REGISTRY_LOGIN = "registry_login"
    RENAME_DATA_DISK = "rename_data_disk"


class IssueType(StrEnum):
    """IssueType type.

    This is an incomplete list. Supervisor regularly adds new types of issues as they
    are discovered. Therefore when returning an issue, it may have a type that is not
    in this list parsed as strings on older versions of the client.
    """

    BOOT_FAIL = "boot_fail"
    CORRUPT_DOCKER = "corrupt_docker"
    CORRUPT_REPOSITORY = "corrupt_repository"
    CORRUPT_FILESYSTEM = "corrupt_filesystem"
    DEPRECATED_ADDON = "deprecated_addon"
    DEPRECATED_ARCH_ADDON = "deprecated_arch_addon"
    DETACHED_ADDON_MISSING = "detached_addon_missing"
    DETACHED_ADDON_REMOVED = "detached_addon_removed"
    DEVICE_ACCESS_MISSING = "device_access_missing"
    DISABLED_DATA_DISK = "disabled_data_disk"
    DISK_LIFETIME = "disk_lifetime"
    DNS_LOOP = "dns_loop"
    DNS_SERVER_FAILED = "dns_server_failed"
    DNS_SERVER_IPV6_ERROR = "dns_server_ipv6_error"
    DOCKER_CONFIG = "docker_config"
    DOCKER_RATELIMIT = "docker_ratelimit"
    FATAL_ERROR = "fatal_error"
    FREE_SPACE = "free_space"
    IPV4_CONNECTION_PROBLEM = "ipv4_connection_problem"
    MISSING_IMAGE = "missing_image"
    MOUNT_FAILED = "mount_failed"
    MULTIPLE_DATA_DISKS = "multiple_data_disks"
    NO_CURRENT_BACKUP = "no_current_backup"
    NTP_SYNC_FAILED = "ntp_sync_failed"
    PWNED = "pwned"
    REBOOT_REQUIRED = "reboot_required"
    SECURITY = "security"
    UPDATE_FAILED = "update_failed"
    UPDATE_ROLLBACK = "update_rollback"


class UnsupportedReason(StrEnum):
    """UnsupportedReason type.

    This is an incomplete list. Supervisor regularly adds new unsupported
    reasons as they are discovered. Therefore when returning a list of unsupported
    reasons, some may not be in this list parsed as strings on older versions of the
    client.
    """

    APPARMOR = "apparmor"
    CGROUP_VERSION = "cgroup_version"
    CONNECTIVITY_CHECK = "connectivity_check"
    DBUS = "dbus"
    DNS_SERVER = "dns_server"
    DOCKER_CONFIGURATION = "docker_configuration"
    DOCKER_VERSION = "docker_version"
    HOME_ASSISTANT_CORE_VERSION = "home_assistant_core_version"
    JOB_CONDITIONS = "job_conditions"
    LXC = "lxc"
    NETWORK_MANAGER = "network_manager"
    OS = "os"
    OS_AGENT = "os_agent"
    OS_VERSION = "os_version"
    PRIVILEGED = "privileged"
    RESTART_POLICY = "restart_policy"
    SOFTWARE = "software"
    SUPERVISOR_VERSION = "supervisor_version"
    SYSTEMD = "systemd"
    SYSTEMD_JOURNAL = "systemd_journal"
    SYSTEMD_RESOLVED = "systemd_resolved"
    VIRTUALIZATION_IMAGE = "virtualization_image"


class UnhealthyReason(StrEnum):
    """UnhealthyReason type.

    This is an incomplete list. Supervisor regularly adds new unhealthy
    reasons as they are discovered. Therefore when returning a list of unhealthy
    reasons, some may not be in this list parsed as strings on older versions of the
    client.
    """

    DOCKER = "docker"
    DUPLICATE_OS_INSTALLATION = "duplicate_os_installation"
    OSERROR_BAD_MESSAGE = "oserror_bad_message"
    PRIVILEGED = "privileged"
    SETUP = "setup"
    SUPERVISOR = "supervisor"
    UNTRUSTED = "untrusted"


class ContextType(StrEnum):
    """ContextType type."""

    ADDON = "addon"
    CORE = "core"
    DNS_SERVER = "dns_server"
    MOUNT = "mount"
    OS = "os"
    PLUGIN = "plugin"
    SUPERVISOR = "supervisor"
    STORE = "store"
    SYSTEM = "system"


class CheckType(StrEnum):
    """CheckType type.

    This is an incomplete list. Supervisor regularly adds new checks as they are
    discovered. Therefore when returning a list of checks, some may have a type that is
    not in this list parsed as strings on older versions of the client.
    """

    ADDON_PWNED = "addon_pwned"
    BACKUPS = "backups"
    DEPRECATED_ADDON = "deprecated_addon"
    DEPRECATED_ARCH_ADDON = "deprecated_arch_addon"
    DETACHED_ADDON_MISSING = "detached_addon_missing"
    DETACHED_ADDON_REMOVED = "detached_addon_removed"
    DISABLED_DATA_DISK = "disabled_data_disk"
    DNS_SERVER_IPV6 = "dns_server_ipv6"
    DNS_SERVER = "dns_server"
    DOCKER_CONFIG = "docker_config"
    FREE_SPACE = "free_space"
    MULTIPLE_DATA_DISKS = "multiple_data_disks"
    NETWORK_INTERFACE_IPV4 = "network_interface_ipv4"


# --- OBJECTS ----


@dataclass(frozen=True, slots=True)
class Suggestion(ResponseData):
    """Suggestion model."""

    type: SuggestionType | str
    context: ContextType
    reference: str | None
    uuid: UUID
    auto: bool


@dataclass(frozen=True, slots=True)
class Issue(ResponseData):
    """Issue model."""

    type: IssueType | str
    context: ContextType
    reference: str | None
    uuid: UUID


@dataclass(frozen=True, slots=True)
class Check(ResponseData):
    """Check model."""

    enabled: bool
    slug: CheckType | str


@dataclass(frozen=True, slots=True)
class SuggestionsList(ResponseData):
    """SuggestionsList model."""

    suggestions: list[Suggestion]


@dataclass(frozen=True, slots=True)
class ResolutionInfo(SuggestionsList, ResponseData):
    """ResolutionInfo model."""

    unsupported: list[UnsupportedReason | str]
    unhealthy: list[UnhealthyReason | str]
    issues: list[Issue]
    checks: list[Check]


@dataclass(frozen=True, slots=True)
class CheckOptions(Options):
    """CheckOptions model."""

    enabled: bool | None = None



"""Store client for supervisor."""

from typing import Any

from .client import _SupervisorComponentClient
from .const import ResponseType
from .models.addons import (
    Repository,
    StoreAddon,
    StoreAddonComplete,
    StoreAddonInstall,
    StoreAddonsList,
    StoreAddonUpdate,
    StoreAddRepository,
    StoreInfo,
)


class StoreClient(_SupervisorComponentClient):
    """Handles store access in Supervisor."""

    async def info(self) -> StoreInfo:
        """Get store info."""
        result = await self._client.get("store")
        return StoreInfo.from_dict(result.data)

    async def addons_list(self) -> list[StoreAddon]:
        """Get list of store addons."""
        result = await self._client.get("store/addons")
        return StoreAddonsList.from_dict(result.data).addons

    async def addon_info(self, addon: str) -> StoreAddonComplete:
        """Get store addon info."""
        result = await self._client.get(f"store/addons/{addon}")
        return StoreAddonComplete.from_dict(result.data)

    async def addon_changelog(self, addon: str) -> str:
        """Get addon changelog."""
        result = await self._client.get(
            f"store/addons/{addon}/changelog", response_type=ResponseType.TEXT
        )
        return result.data

    async def addon_documentation(self, addon: str) -> str:
        """Get addon documentation."""
        result = await self._client.get(
            f"store/addons/{addon}/documentation", response_type=ResponseType.TEXT
        )
        return result.data

    async def addon_availability(self, addon: str) -> None:
        """Determine if latest version of addon can be installed on this system.

        No return means it can be. If not, raises one of the following errors:
        - AddonUnavailableHomeAssistantVersionError
        - AddonUnavailableArchitectureError
        - AddonUnavailableMachineTypeError

        If Supervisor adds a new reason an add-on can be restricted from being
        installed on some systems in the future, older versions of this client
        will raise the generic SupervisorBadRequestError for that reason.
        """
        await self._client.get(
            f"store/addons/{addon}/availability", response_type=ResponseType.NONE
        )

    async def install_addon(
        self, addon: str, options: StoreAddonInstall | None = None
    ) -> None:
        """Install an addon.

        Supervisor does an availability check before install. If the addon
        cannot be installed on this system it will raise one of those errors
        shown in the `addon_availability` method.
        """
        # Must disable timeout if API call waits for install to complete
        kwargs: dict[str, Any] = {}
        if not options or not options.background:
            kwargs["timeout"] = None

        await self._client.post(
            f"store/addons/{addon}/install",
            json=options.to_dict() if options else None,
            **kwargs,
        )

    async def update_addon(
        self, addon: str, options: StoreAddonUpdate | None = None
    ) -> None:
        """Update an addon to latest version.

        Supervisor does an availability check before update. If the new version
        cannot be installed on this system it will raise one of those errors
        shown in the `addon_availability` method.
        """
        # Must disable timeout if API call waits for update to complete
        kwargs: dict[str, Any] = {}
        if not options or not options.background:
            kwargs["timeout"] = None

        await self._client.post(
            f"store/addons/{addon}/update",
            json=options.to_dict() if options else None,
            **kwargs,
        )

    async def reload(self) -> None:
        """Reload the store."""
        await self._client.post("store/reload")

    async def repositories_list(self) -> list[Repository]:
        """Get list of repositories."""
        result = await self._client.get("store/repositories")
        # This API is inconsistent with Supervisor's API model, data should be
        # a dictionary with a "repositories" field. It would break the CLI like
        # this but the CLI doesn't use it so it went unnoticed.
        return [Repository.from_dict(repo) for repo in result.data]

    async def repository_info(self, repository: str) -> Repository:
        """Get repository info."""
        result = await self._client.get(f"store/repositories/{repository}")
        return Repository.from_dict(result.data)

    async def add_repository(self, options: StoreAddRepository) -> None:
        """Add a repository to the store."""
        await self._client.post("store/repositories", json=options.to_dict())

    async def remove_repository(self, repository: str) -> None:
        """Remove a repository from the store."""
        await self._client.delete(f"store/repositories/{repository}")

    async def repair_repository(self, repository: str) -> None:
        """Repair/reset an addon repository in the store."""
        await self._client.post(f"store/repositories/{repository}/repair")

    # Omitted for now - Icon/Logo endpoints



"""Internal client for making requests and managing session with Supervisor."""

from dataclasses import dataclass, field
from http import HTTPMethod, HTTPStatus
from importlib import metadata
from typing import Any

from aiohttp import (
    ClientError,
    ClientResponse,
    ClientResponseError,
    ClientSession,
    ClientTimeout,
)
from multidict import MultiDict
from yarl import URL

from .const import DEFAULT_TIMEOUT, ResponseType
from .exceptions import (
    ERROR_KEYS,
    SupervisorAuthenticationError,
    SupervisorBadRequestError,
    SupervisorConnectionError,
    SupervisorError,
    SupervisorForbiddenError,
    SupervisorNotFoundError,
    SupervisorResponseError,
    SupervisorServiceUnavailableError,
    SupervisorTimeoutError,
)
from .models.base import Response, ResultType
from .utils.aiohttp import ChunkAsyncStreamIterator

VERSION = metadata.version(__package__)


def is_json(response: ClientResponse, *, raise_on_fail: bool = False) -> bool:
    """Check if response is json according to Content-Type."""
    content_type = response.headers.get("Content-Type", "")
    if "application/json" not in content_type:
        if raise_on_fail:
            raise SupervisorResponseError(
                "Unexpected response received from supervisor when expecting"
                f"JSON. Status: {response.status}, content type: {content_type}",
            )
        return False
    return True


@dataclass(slots=True)
class _SupervisorClient:
    """Main class for handling connections with Supervisor."""

    api_host: str
    token: str
    session: ClientSession | None = None
    _close_session: bool = field(default=False, init=False)

    async def _raise_on_status(self, response: ClientResponse) -> None:
        """Raise appropriate exception on status."""
        if response.status >= HTTPStatus.BAD_REQUEST.value:
            exc_type: type[SupervisorError] = SupervisorError
            match response.status:
                case HTTPStatus.BAD_REQUEST:
                    exc_type = SupervisorBadRequestError
                case HTTPStatus.UNAUTHORIZED:
                    exc_type = SupervisorAuthenticationError
                case HTTPStatus.FORBIDDEN:
                    exc_type = SupervisorForbiddenError
                case HTTPStatus.NOT_FOUND:
                    exc_type = SupervisorNotFoundError
                case HTTPStatus.SERVICE_UNAVAILABLE:
                    exc_type = SupervisorServiceUnavailableError

            if is_json(response):
                result = Response.from_json(await response.text())
                if result.error_key in ERROR_KEYS:
                    exc_type = ERROR_KEYS[result.error_key]
                raise exc_type(
                    result.message,
                    result.extra_fields,
                    result.job_id,
                )
            raise exc_type()

    async def _request(
        self,
        method: HTTPMethod,
        uri: str,
        *,
        params: dict[str, str] | MultiDict[str] | None,
        response_type: ResponseType,
        json: dict[str, Any] | None = None,
        data: Any = None,
        timeout: ClientTimeout | None = DEFAULT_TIMEOUT,
    ) -> Response:
        """Handle a request to Supervisor."""
        try:
            url = URL(self.api_host).joinpath(uri)
        except ValueError as err:
            raise SupervisorError from err

        # This check is to make sure the normalized URL string is the same as the URL
        # string that was passed in. If they are different, then the passed in uri
        # contained characters that were removed by the normalization
        # such as ../../../../etc/passwd
        if not url.raw_path.endswith(uri):
            raise SupervisorError(f"Invalid request {uri}")

        match response_type:
            case ResponseType.TEXT:
                accept = "text/plain, */*"
            case _:
                accept = "application/json, text/plain, */*"

        headers = {
            "User-Agent": f"AioHASupervisor/{VERSION}",
            "Accept": accept,
            "Authorization": f"Bearer {self.token}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            response = await self.session.request(
                method.value,
                url,
                timeout=timeout,
                headers=headers,
                params=params,
                json=json,
                data=data,
            )
            await self._raise_on_status(response)
            match response_type:
                case ResponseType.JSON:
                    is_json(response, raise_on_fail=True)
                    return Response.from_json(await response.text())
                case ResponseType.TEXT:
                    return Response(ResultType.OK, await response.text())
                case ResponseType.STREAM:
                    return Response(
                        ResultType.OK, ChunkAsyncStreamIterator(response.content)
                    )
                case _:
                    return Response(ResultType.OK)

        except (UnicodeDecodeError, ClientResponseError) as err:
            raise SupervisorResponseError(
                "Unusable response received from Supervisor, check logs",
            ) from err
        except TimeoutError as err:
            raise SupervisorTimeoutError("Timeout connecting to Supervisor") from err
        except ClientError as err:
            raise SupervisorConnectionError(
                "Error occurred connecting to supervisor",
            ) from err

    async def get(
        self,
        uri: str,
        *,
        params: dict[str, str] | MultiDict[str] | None = None,
        response_type: ResponseType = ResponseType.JSON,
        timeout: ClientTimeout | None = DEFAULT_TIMEOUT,
    ) -> Response:
        """Handle a GET request to Supervisor."""
        return await self._request(
            HTTPMethod.GET,
            uri,
            params=params,
            response_type=response_type,
            timeout=timeout,
        )

    async def post(
        self,
        uri: str,
        *,
        params: dict[str, str] | MultiDict[str] | None = None,
        response_type: ResponseType = ResponseType.NONE,
        json: dict[str, Any] | None = None,
        data: Any = None,
        timeout: ClientTimeout | None = DEFAULT_TIMEOUT,
    ) -> Response:
        """Handle a POST request to Supervisor."""
        return await self._request(
            HTTPMethod.POST,
            uri,
            params=params,
            response_type=response_type,
            json=json,
            data=data,
            timeout=timeout,
        )

    async def put(
        self,
        uri: str,
        *,
        params: dict[str, str] | MultiDict[str] | None = None,
        json: dict[str, Any] | None = None,
        timeout: ClientTimeout | None = DEFAULT_TIMEOUT,
    ) -> Response:
        """Handle a PUT request to Supervisor."""
        return await self._request(
            HTTPMethod.PUT,
            uri,
            params=params,
            response_type=ResponseType.NONE,
            json=json,
            timeout=timeout,
        )

    async def delete(
        self,
        uri: str,
        *,
        params: dict[str, str] | MultiDict[str] | None = None,
        json: dict[str, Any] | None = None,
        timeout: ClientTimeout | None = DEFAULT_TIMEOUT,
    ) -> Response:
        """Handle a DELETE request to Supervisor."""
        return await self._request(
            HTTPMethod.DELETE,
            uri,
            params=params,
            response_type=ResponseType.NONE,
            json=json,
            timeout=timeout,
        )

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()


class _SupervisorComponentClient:
    """Common ancestor for all component clients of supervisor."""

    def __init__(self, client: _SupervisorClient) -> None:
        """Initialize sub module with client for API calls."""
        self._client = client



"""Test addons supervisor client."""

from ipaddress import IPv4Address

from aioresponses import aioresponses
from yarl import URL

from aiohasupervisor import SupervisorClient
from aiohasupervisor.models import (
    AddonBoot,
    AddonsOptions,
    AddonsRebuild,
    AddonsSecurityOptions,
    AddonStage,
    AddonStartup,
    AddonState,
    AddonsUninstall,
    Capability,
    InstalledAddonComplete,
    StoreAddonComplete,
    SupervisorRole,
)
from aiohasupervisor.models.base import Response

from . import load_fixture
from .const import SUPERVISOR_URL


async def test_addons_list(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addons list API."""
    responses.get(
        f"{SUPERVISOR_URL}/addons", status=200, body=load_fixture("addons_list.json")
    )
    addons = await supervisor_client.addons.list()
    assert addons[0].name == "Terminal & SSH"
    assert addons[0].slug == "core_ssh"
    assert addons[0].icon is True
    assert addons[0].logo is True
    assert addons[0].state == AddonState.STARTED
    assert addons[0].stage == AddonStage.STABLE
    assert addons[1].slug == "a0d7b954_vscode"


async def test_addons_info(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addons info API."""
    responses.get(
        f"{SUPERVISOR_URL}/addons/core_ssh/info",
        status=200,
        body=load_fixture("addons_info.json"),
    )
    addon = await supervisor_client.addons.addon_info("core_ssh")
    assert addon.name == "Terminal & SSH"
    assert addon.slug == "core_ssh"
    assert addon.documentation is True
    assert addon.changelog is True
    assert addon.watchdog is False
    assert addon.auto_update is False
    assert addon.ip_address == IPv4Address("172.30.33.0")
    assert Capability.NET_RAW in addon.privileged
    assert "not_real" in addon.privileged
    assert addon.supervisor_api is True
    assert addon.supervisor_role == SupervisorRole.MANAGER
    assert addon.system_managed is False
    assert addon.system_managed_config_entry is None
    assert addon.startup == AddonStartup.SERVICES


async def test_addons_uninstall(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test uninstall addon API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/uninstall", status=200)
    assert (
        await supervisor_client.addons.uninstall_addon(
            "core_ssh", AddonsUninstall(remove_config=True)
        )
        is None
    )
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/uninstall"))
    }


async def test_addons_start(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test start addon API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/start", status=200)
    assert await supervisor_client.addons.start_addon("core_ssh") is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/start"))
    }


async def test_addons_stop(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test stop addon API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/stop", status=200)
    assert await supervisor_client.addons.stop_addon("core_ssh") is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/stop"))
    }


async def test_addons_restart(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test restart addon API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/restart", status=200)
    assert await supervisor_client.addons.restart_addon("core_ssh") is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/restart"))
    }


async def test_addons_options(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addon options API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/options", status=200)
    assert (
        await supervisor_client.addons.set_addon_options(
            "core_ssh",
            AddonsOptions(
                config=None,
                boot=AddonBoot.AUTO,
                network={"22/tcp": 22, "1234/tcp": None},
                watchdog=True,
            ),
        )
        is None
    )
    assert len(responses.requests) == 1
    assert (
        request := responses.requests[
            ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/options"))
        ]
    )
    assert request[0].kwargs["json"] == {
        "options": None,
        "boot": "auto",
        "network": {"22/tcp": 22, "1234/tcp": None},
        "watchdog": True,
    }


async def test_addons_config_validate(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test config validate API."""
    responses.post(
        f"{SUPERVISOR_URL}/addons/core_ssh/options/validate",
        status=200,
        body=load_fixture("addons_config_validate.json"),
    )
    validate = await supervisor_client.addons.addon_config_validate(
        "core_ssh", {"bad": "config"}
    )
    assert validate.message == (
        "Missing option 'server' in root in Terminal & SSH"
        " (core_ssh). Got {'bad': 'config'}"
    )
    assert validate.valid is False
    assert validate.pwned is False


async def test_addons_config(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test get addon config API."""
    responses.get(
        f"{SUPERVISOR_URL}/addons/core_ssh/options/config",
        status=200,
        body=load_fixture("addons_options_config.json"),
    )
    config = await supervisor_client.addons.addon_config("core_ssh")
    assert config["authorized_keys"] == []
    assert config["password"] == ""


async def test_addons_rebuild(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test rebuild addon API."""
    responses.post(f"{SUPERVISOR_URL}/addons/local_example/rebuild", status=200)
    assert await supervisor_client.addons.rebuild_addon("local_example") is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/addons/local_example/rebuild"))
    }


async def test_addons_rebuild_with_force(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test rebuild addon API with force option."""
    responses.post(f"{SUPERVISOR_URL}/addons/local_example/rebuild", status=200)
    assert (
        await supervisor_client.addons.rebuild_addon(
            "local_example", AddonsRebuild(force=True)
        )
        is None
    )

    request_key = ("POST", URL(f"{SUPERVISOR_URL}/addons/local_example/rebuild"))
    assert request_key in responses.requests

    request_call = responses.requests[request_key][0]
    assert request_call.kwargs["json"] == {"force": True}


async def test_addons_stdin(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addon stdin API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/stdin", status=200)
    assert (
        await supervisor_client.addons.write_addon_stdin("core_ssh", b"hello world")
        is None
    )
    assert len(responses.requests) == 1
    assert (
        request := responses.requests[
            ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/stdin"))
        ]
    )
    assert request[0].kwargs["data"] == b"hello world"


async def test_addons_security(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addon security API."""
    responses.post(f"{SUPERVISOR_URL}/addons/core_ssh/security", status=200)
    assert (
        await supervisor_client.addons.set_addon_security(
            "core_ssh", AddonsSecurityOptions(protected=True)
        )
        is None
    )
    assert len(responses.requests) == 1
    assert (
        request := responses.requests[
            ("POST", URL(f"{SUPERVISOR_URL}/addons/core_ssh/security"))
        ]
    )
    assert request[0].kwargs["json"] == {"protected": True}


async def test_addons_stats(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test addon stats API."""
    responses.get(
        f"{SUPERVISOR_URL}/addons/core_ssh/stats",
        status=200,
        body=load_fixture("addon_stats.json"),
    )
    stats = await supervisor_client.addons.addon_stats("core_ssh")
    assert stats.cpu_percent == 0
    assert stats.memory_usage == 24588288
    assert stats.network_rx == 1717120021


async def test_addons_serialize_by_alias() -> None:
    """Test serializing addons by alias."""
    response = Response.from_json(load_fixture("store_addon_info.json"))
    store_addon_info = StoreAddonComplete.from_dict(response.data)

    assert store_addon_info.supervisor_api is False
    assert (store_addon_info.to_dict())["supervisor_api"] is False
    assert (store_addon_info.to_dict(by_alias=True))["hassio_api"] is False

    response = Response.from_json(load_fixture("addons_info.json"))
    addon_info = InstalledAddonComplete.from_dict(response.data)

    assert addon_info.supervisor_api is True
    assert (addon_info.to_dict())["supervisor_api"] is True
    assert (addon_info.to_dict(by_alias=True))["hassio_api"] is True



"""Base types and internal models."""

from abc import ABC
from dataclasses import dataclass
from enum import StrEnum
from typing import Any, Literal

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin


class SentinelMeta(type):
    """Metaclass for sentinel to improve representation and make falsy.

    Credit to https://stackoverflow.com/a/69243488 .
    """

    def __repr__(cls) -> str:
        """Represent class more like an enum."""
        return f"<{cls.__name__}>"

    def __bool__(cls) -> Literal[False]:
        """Return false as a sentinel is akin to an empty value."""
        return False


class DEFAULT(metaclass=SentinelMeta):
    """Sentinel for default value when None is valid."""


class RequestConfig(BaseConfig):
    """Default Mashumaro config for all request models."""

    omit_default = True


@dataclass(frozen=True)
class Request(ABC, DataClassDictMixin):
    """Omit default in requests to allow Supervisor to set default.

    If None is a valid value, the default value should be the sentinel
    DEFAULT for optional fields.
    """

    class Config(RequestConfig):
        """Mashumaro config."""


@dataclass(frozen=True)
class Options(ABC, DataClassDictMixin):
    """Superclass for Options models to ensure a field is present.

    All fields should be optional. If None is a valid value, use the DEFAULT
    sentinel. Client should only pass changed fields to Supervisor.
    """

    def __post_init__(self) -> None:
        """Validate at least one field is present."""
        if not self.to_dict():
            raise ValueError("At least one field must have a value")

    class Config(RequestConfig):
        """Mashumaro config."""


@dataclass(frozen=True)
class ResponseData(ABC, DataClassDictMixin):
    """Superclass for all response data objects."""


class ResultType(StrEnum):
    """ResultType type."""

    OK = "ok"
    ERROR = "error"


@dataclass(frozen=True, slots=True)
class Response(DataClassORJSONMixin):
    """Response model for all JSON based endpoints."""

    result: ResultType
    data: Any | None = None
    message: str | None = None
    job_id: str | None = None
    error_key: str | None = None
    extra_fields: dict[str, Any] | None = None


@dataclass(frozen=True, slots=True)
class ContainerStats(ResponseData):
    """ContainerStats model."""

    cpu_percent: float
    memory_usage: int
    memory_limit: int
    memory_percent: float
    network_rx: int
    network_tx: int
    blk_read: int
    blk_write: int



"""OS client for supervisor."""

from .client import _SupervisorComponentClient
from .models.os import (
    DataDisk,
    DataDiskList,
    GreenInfo,
    GreenOptions,
    MigrateDataOptions,
    OSInfo,
    OSUpdate,
    SetBootSlotOptions,
    SwapInfo,
    SwapOptions,
    YellowInfo,
    YellowOptions,
)


class OSClient(_SupervisorComponentClient):
    """Handles OS access in supervisor."""

    async def info(self) -> OSInfo:
        """Get OS info."""
        result = await self._client.get("os/info")
        return OSInfo.from_dict(result.data)

    async def update(self, options: OSUpdate | None = None) -> None:
        """Update OS."""
        await self._client.post(
            "os/update", json=options.to_dict() if options else None, timeout=None
        )

    async def swap_info(self) -> SwapInfo:
        """Get swap settings."""
        result = await self._client.get("os/config/swap")
        return SwapInfo.from_dict(result.data)

    async def set_swap_options(self, options: SwapOptions) -> None:
        """Set swap settings."""
        await self._client.post("os/config/swap", json=options.to_dict())

    async def config_sync(self) -> None:
        """Trigger config reload on OS."""
        await self._client.post("os/config/sync")

    async def migrate_data(self, options: MigrateDataOptions) -> None:
        """Migrate data to new data disk and reboot."""
        await self._client.post("os/datadisk/move", json=options.to_dict())

    async def list_data_disks(self) -> list[DataDisk]:
        """Get all data disks."""
        result = await self._client.get("os/datadisk/list")
        return DataDiskList.from_dict(result.data).disks

    async def wipe_data(self) -> None:
        """Trigger data disk wipe on host and reboot."""
        await self._client.post("os/datadisk/wipe")

    async def set_boot_slot(self, options: SetBootSlotOptions) -> None:
        """Change active boot slot on host and reboot."""
        await self._client.post("os/boot-slot", json=options.to_dict())

    async def green_info(self) -> GreenInfo:
        """Get info for green board (if in use)."""
        result = await self._client.get("os/boards/green")
        return GreenInfo.from_dict(result.data)

    async def set_green_options(self, options: GreenOptions) -> None:
        """Set options for green board (if in use)."""
        await self._client.post("os/boards/green", json=options.to_dict())

    async def yellow_info(self) -> YellowInfo:
        """Get info for yellow board (if in use)."""
        result = await self._client.get("os/boards/yellow")
        return YellowInfo.from_dict(result.data)

    async def set_yellow_options(self, options: YellowOptions) -> None:
        """Set options for yellow board (if in use)."""
        await self._client.post("os/boards/yellow", json=options.to_dict())



"""Models for Supervisor mounts."""

from abc import ABC
from dataclasses import dataclass, field
from enum import StrEnum
from pathlib import PurePath
from typing import Literal

from .base import Request, ResponseData

# --- ENUMS ----


class MountType(StrEnum):
    """MountType type."""

    CIFS = "cifs"
    NFS = "nfs"


class MountUsage(StrEnum):
    """MountUsage type."""

    BACKUP = "backup"
    MEDIA = "media"
    SHARE = "share"


class MountState(StrEnum):
    """MountState type."""

    ACTIVE = "active"
    ACTIVATING = "activating"
    DEACTIVATING = "deactivating"
    FAILED = "failed"
    INACTIVE = "inactive"
    MAINTENANCE = "maintenance"
    RELOADING = "reloading"


class MountCifsVersion(StrEnum):
    """Mount CIFS version."""

    LEGACY_1_0 = "1.0"
    LEGACY_2_0 = "2.0"


# --- OBJECTS ----


@dataclass(frozen=True)
class Mount(ABC):
    """Mount ABC type."""

    usage: MountUsage
    server: str
    port: int | None = field(kw_only=True, default=None)


@dataclass(frozen=True)
class CIFSMount(ABC):
    """CIFSMount ABC type."""

    share: str
    version: MountCifsVersion | None = field(kw_only=True, default=None)


@dataclass(frozen=True)
class NFSMount(ABC):
    """NFSMount ABC type."""

    path: PurePath


@dataclass(frozen=True)
class MountResponse(ABC):
    """MountResponse model."""

    name: str
    read_only: bool
    state: MountState | None
    user_path: PurePath | None


@dataclass(frozen=True)
class MountRequest(ABC):  # noqa: B024
    """MountRequest model."""

    read_only: bool | None = field(kw_only=True, default=None)


@dataclass(frozen=True, slots=True)
class CIFSMountResponse(Mount, MountResponse, CIFSMount, ResponseData):
    """CIFSMountResponse model."""

    type: Literal[MountType.CIFS]


@dataclass(frozen=True, slots=True)
class NFSMountResponse(Mount, MountResponse, NFSMount, ResponseData):
    """NFSMountResponse model."""

    type: Literal[MountType.NFS]


@dataclass(frozen=True, slots=True)
class CIFSMountRequest(Mount, MountRequest, CIFSMount, Request):
    """CIFSMountRequest model."""

    type: Literal[MountType.CIFS] = field(init=False, default=MountType.CIFS)
    username: str | None = field(kw_only=True, default=None)
    password: str | None = field(kw_only=True, default=None)


@dataclass(frozen=True, slots=True)
class NFSMountRequest(Mount, MountRequest, NFSMount, Request):
    """NFSMountRequest model."""

    type: Literal[MountType.NFS] = field(init=False, default=MountType.NFS)


@dataclass(frozen=True, slots=True)
class MountsInfo(ResponseData):
    """MountsInfo model."""

    default_backup_mount: str | None
    mounts: list[CIFSMountResponse | NFSMountResponse]


@dataclass(frozen=True, slots=True)
class MountsOptions(Request):
    """MountsOptions model."""

    default_backup_mount: str | None



"""Test discovery supervisor client."""

from uuid import UUID

from aioresponses import aioresponses
from yarl import URL

from aiohasupervisor import SupervisorClient
from aiohasupervisor.models import DiscoveryConfig

from . import load_fixture
from .const import SUPERVISOR_URL


async def test_discovery_list(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test Discovery list API."""
    responses.get(
        f"{SUPERVISOR_URL}/discovery",
        status=200,
        body=load_fixture("discovery_list.json"),
    )
    disc_list = await supervisor_client.discovery.list()
    assert disc_list[0].addon == "core_mosquitto"
    assert disc_list[0].service == "mqtt"
    assert disc_list[0].uuid.hex == "889ca604cff84004894e53d181655b3a"
    assert disc_list[0].config["host"] == "core-mosquitto"


async def test_get_discovery(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test Discovery get API."""
    uuid = UUID("889ca604cff84004894e53d181655b3a")
    responses.get(
        f"{SUPERVISOR_URL}/discovery/{uuid.hex}",
        status=200,
        body=load_fixture("discovery_get.json"),
    )
    discovery = await supervisor_client.discovery.get(uuid)
    assert discovery.addon == "core_mosquitto"
    assert discovery.service == "mqtt"
    assert discovery.uuid == uuid
    assert discovery.config["host"] == "core-mosquitto"
    assert discovery.config["port"] == 1883
    assert discovery.config["ssl"] is False


async def test_delete_discovery(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test Discovery delete API."""
    uuid = UUID("889ca604cff84004894e53d181655b3a")
    responses.delete(f"{SUPERVISOR_URL}/discovery/{uuid.hex}", status=200)
    assert await supervisor_client.discovery.delete(uuid) is None
    assert responses.requests.keys() == {
        ("DELETE", URL(f"{SUPERVISOR_URL}/discovery/{uuid.hex}"))
    }


async def test_set_discovery(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test Discovery set API."""
    responses.post(
        f"{SUPERVISOR_URL}/discovery",
        status=200,
        body=load_fixture("discovery_set.json"),
    )
    assert await supervisor_client.discovery.set(
        DiscoveryConfig(service="mqtt", config={})
    ) == UUID("889ca604cff84004894e53d181655b3a")



"""Exceptions from supervisor client."""

from abc import ABC
from collections.abc import Callable
from typing import Any


class SupervisorError(Exception):
    """Generic exception."""

    error_key: str | None = None

    def __init__(
        self,
        message: str | None = None,
        extra_fields: dict[str, Any] | None = None,
        job_id: str | None = None,
    ) -> None:
        """Initialize exception."""
        if message is not None:
            super().__init__(message)
        else:
            super().__init__()

        self.job_id = job_id
        self.extra_fields = extra_fields


ERROR_KEYS: dict[str, type[SupervisorError]] = {}


def error_key(
    key: str,
) -> Callable[[type[SupervisorError]], type[SupervisorError]]:
    """Store exception in keyed error map."""

    def wrap(cls: type[SupervisorError]) -> type[SupervisorError]:
        ERROR_KEYS[key] = cls
        cls.error_key = key
        return cls

    return wrap


class SupervisorConnectionError(SupervisorError, ConnectionError):
    """Unknown error connecting to supervisor."""


class SupervisorTimeoutError(SupervisorError, TimeoutError):
    """Timeout connecting to supervisor."""


class SupervisorBadRequestError(SupervisorError):
    """Invalid request made to supervisor."""


class SupervisorAuthenticationError(SupervisorError):
    """Invalid authentication sent to supervisor."""


class SupervisorForbiddenError(SupervisorError):
    """Client is not allowed to take the action requested."""


class SupervisorNotFoundError(SupervisorError):
    """Requested resource does not exist."""


class SupervisorServiceUnavailableError(SupervisorError):
    """Cannot complete request because a required service is unavailable."""


class SupervisorResponseError(SupervisorError):
    """Unusable response received from Supervisor with the wrong type or encoding."""


class AddonNotSupportedError(SupervisorError, ABC):
    """Addon is not supported on this system."""


@error_key("addon_not_supported_architecture_error")
class AddonNotSupportedArchitectureError(AddonNotSupportedError):
    """Addon is not supported on this system due to its architecture."""


@error_key("addon_not_supported_machine_type_error")
class AddonNotSupportedMachineTypeError(AddonNotSupportedError):
    """Addon is not supported on this system due to its machine type."""


@error_key("addon_not_supported_home_assistant_version_error")
class AddonNotSupportedHomeAssistantVersionError(AddonNotSupportedError):
    """Addon is not supported on this system due to its version of Home Assistant."""



"""Home Assistant client for supervisor."""

from .client import _SupervisorComponentClient
from .models.homeassistant import (
    HomeAssistantInfo,
    HomeAssistantOptions,
    HomeAssistantRebuildOptions,
    HomeAssistantRestartOptions,
    HomeAssistantStats,
    HomeAssistantStopOptions,
    HomeAssistantUpdateOptions,
)


class HomeAssistantClient(_SupervisorComponentClient):
    """Handles Home Assistant access in supervisor."""

    async def info(self) -> HomeAssistantInfo:
        """Get Home Assistant info."""
        result = await self._client.get("core/info")
        return HomeAssistantInfo.from_dict(result.data)

    async def stats(self) -> HomeAssistantStats:
        """Get Home Assistant stats."""
        result = await self._client.get("core/stats")
        return HomeAssistantStats.from_dict(result.data)

    async def set_options(self, options: HomeAssistantOptions) -> None:
        """Set Home Assistant options."""
        await self._client.post("core/options", json=options.to_dict())

    async def update(self, options: HomeAssistantUpdateOptions | None = None) -> None:
        """Update Home Assistant."""
        await self._client.post(
            "core/update", json=options.to_dict() if options else None, timeout=None
        )

    async def restart(self, options: HomeAssistantRestartOptions | None = None) -> None:
        """Restart Home Assistant."""
        await self._client.post(
            "core/restart", json=options.to_dict() if options else None
        )

    async def stop(self, options: HomeAssistantStopOptions | None = None) -> None:
        """Stop Home Assistant."""
        await self._client.post(
            "core/stop", json=options.to_dict() if options else None
        )

    async def start(self) -> None:
        """Start Home Assistant."""
        await self._client.post("core/start")

    async def check_config(self) -> None:
        """Check Home Assistant config."""
        await self._client.post("core/check")

    async def rebuild(self, options: HomeAssistantRebuildOptions | None = None) -> None:
        """Rebuild Home Assistant."""
        await self._client.post(
            "core/rebuild", json=options.to_dict() if options else None
        )
